{
  Roland Potter - Woodstock Computer Solutions
  http://wcsweb.co.uk

  This code is provided for educatonal purposes only
  with no warranty whatsoever.

  Feel free to adapt it and use it in applcations commercial
  or otherwise and open source componenents but not as part
  of commercially distrbuted developer components.

  Metadata objects used as versioned-streamable base
  objects for Data definitions.
}
unit WcsDataObjects;

interface
uses
  Windows, SysUtils, Classes, WCSStreamer;

type
  TWcsEditState =
    (wes_None           // not editing
    ,wes_EditModify     // normal editing - all events fired
    ,wes_EditNoCancel   // normal editing - skip backup, either post or raise exception
    ,wes_EditNoModify   // allows update to fields without setting modify flags (Post/Cancel does nothing)
    );

  TWcsDataObject = class(TObject)
  private
    fParentData: TWcsDataObject;
    fModified : Boolean;
    fUnsaved : Boolean;
    fReadOnly: Boolean;
    fOldValues: TMemoryStream;
    fEditState: TWcsEditState;
    fOnModified: TNotifyEvent;
    fOnAfterPost: TNotifyEvent;
    fOnStateChange: TNotifyEvent;
    fReadStreamVersion,
    fWriteStreamVersion : integer;
    fStringLimit : integer;
    procedure SetModified(const Value: Boolean);
    procedure SetReadOnly(const Value: Boolean);
    procedure SetEditState(const Value: TWcsEditState);
    procedure Backup;
    procedure Restore;
    function GetEditing: Boolean;
    procedure ChildPosted(const ChildData: TWCSDataObject);
    procedure ChildStateChange(const ChildData: TWCSDataObject);
    procedure RaiseNamedError(const Msg: String);
    function GetCurrStreamVersion(const Writing: Boolean): integer;
  protected
    procedure CheckEditing;
    procedure LinkChild(const ChildData: TWCSDataObject);

    function SetByteValue(const Value: Byte; var PrivateVar: Byte): Boolean;
    function SetIntValue(const Value: Longint; var PrivateVar: Longint): Boolean;
    function SetStrValue(const Value: String; var PrivateVar: String): Boolean;
    function SetEnumValue(const Value; var PrivateVar): Boolean;

    // Now obsolete, ONLY used for reading OLD streams!
    procedure StreamStrValueOLD(const Stream: TStream; var Value: String);
    procedure StreamSubValueListOLD(const Stream: TStream; const Values: TList;
      const Limit: integer = MaxInt);
    procedure StreamValuesOLD(const Stream: TStream); Virtual; Abstract;
    procedure CorrectReadStreamVersion(const CorrectVersion: integer);

    procedure StreamSubValues(const Streamer: TWStreamer; var AnObject: TObject);
    procedure StreamSubObject(const Streamer: TWStreamer; const AnObject: TWcsDataObject);

    procedure StreamResolve; virtual;
    function CreateSubData: TWCSDataObject; virtual;
    procedure SetUnsaved(const Value: Boolean); virtual;
    procedure CreateSubClasses; virtual;
    procedure FreeSubClasses; virtual;
    function GetDataObjName: String; virtual;
  public
    constructor Create(const AWriteStreamVersion: integer); virtual;
    destructor Destroy; override;
    procedure Clear(const AndReinitialize: Boolean); virtual;
    procedure StreamValues(const Streamer: TWStreamer); Virtual; Abstract;
    procedure Init; virtual;
    procedure Assign(const Source: TWcsDataObject); virtual;
    procedure StreamIn(const Stream: TStream);
    procedure StreamOut(const Stream: TStream);
    property ReadStreamVersion: integer read fReadStreamVersion;
    property WriteStreamVersion: integer read fWriteStreamVersion;
    property CurrStreamVersion[const Writing: Boolean]: integer read GetCurrStreamVersion;
    // Only call the inherited version when all substreams have been FORCED to re-stream.
    procedure UpdateSubStreams; virtual;
    procedure Edit;
    procedure EditNoCancel;
    function EditNoModify: Boolean;
    procedure Post;
    procedure Cancel;
    // Same as "Post" but with no event triggers!
    // Simply resets the edit state.
    // Used by parent objects before they cancel.
    procedure ResetEdit;

    property Editing: Boolean read GetEditing;
    property EditState:TWcsEditState read fEditState;
    property Modified: Boolean read fModified write SetModified;
    property Unsaved: Boolean read fUnsaved write SetUnsaved;
  // published....
    property ReadOnly: Boolean read fReadOnly write SetReadOnly default False;
    property StringLimit : integer read fStringLimit write fStringLimit default MaxInt;

    property OnModified: TNotifyEvent read fOnModified write fOnModified;
    property OnAfterPost: TNotifyEvent read fOnAfterPost write fOnAfterPost;
    property OnStateChange: TNotifyEvent read fOnStateChange write fOnStateChange;
  end;

implementation

{ TWcsDataObject }

constructor TWcsDataObject.Create(const AWriteStreamVersion: integer);
begin
  fStringLimit := MaxInt;
  fOldValues := TMemoryStream.Create;
  fWriteStreamVersion := AWriteStreamVersion;
  CreateSubClasses;
  Clear(True);
end;

destructor TWcsDataObject.Destroy;
begin
  Clear(False);
  FreeSubClasses;
  fOldValues.Free;
  inherited;
end;

procedure TWcsDataObject.Clear(const AndReinitialize: Boolean);
begin
  UnSaved := False;
  if AndReinitialize then
    Init;
end;

procedure TWcsDataObject.SetModified(const Value: Boolean);
begin
  if (fModified <> Value) and
     (fEditState <> wes_EditNoModify) then
  begin
    fModified := Value;
    if not Value then
      fOldValues.Clear
    else
      if Assigned(fOnModified) then
        fOnModified(Self);
  end;
end;

procedure TWcsDataObject.RaiseNamedError(const Msg: String);
begin
  raise Exception.Create(GetDataObjName+': '+Msg);
end;

procedure TWcsDataObject.SetReadOnly(const Value: Boolean);
begin
  if fReadOnly <> Value then
  begin
    if fModified then
      RaiseNamedError('Cannot set read-only with pending updates');
    fReadOnly := Value;
  end;
end;

procedure TWcsDataObject.CheckEditing;
begin
  if fEditState = wes_None then
    RaiseNamedError('Not in edit mode');
end;

function TWcsDataObject.SetIntValue(const Value: Longint; var PrivateVar: Longint): Boolean;
begin
  Result := (Value <> PrivateVar);
  if Result then
  begin
    CheckEditing;
    PrivateVar := Value;
    Modified := True;
  end;
end;

function TWcsDataObject.SetStrValue(const Value: String; var PrivateVar: String): Boolean;
begin
  Result := (Value <> PrivateVar);
  if Result then
  begin
    CheckEditing;
    PrivateVar := Value;
    Modified := True;
  end;
end;

function TWcsDataObject.SetByteValue(const Value: Byte; var PrivateVar: Byte): Boolean;
begin
  Result := (Value <> PrivateVar);
  if Result then
  begin
    CheckEditing;
    PrivateVar := Value;
    Modified := True;
  end;
end;

function TWcsDataObject.SetEnumValue(const Value; var PrivateVar): Boolean;
begin
  Result := (Byte(Value) <> Byte(PrivateVar));
  if Result then
  begin
    CheckEditing;
    Byte(PrivateVar) := Byte(Value);
    Modified := True;
  end;
end;

{
procedure TWcsDataObject.StreamStreamValue(const Stream: TStream; const Value: TStream;
  const Writing: Boolean);
var
  Len : integer;
begin
  if Writing then
  begin
    Len := Value.Size;
    Stream.Write(Len, 4);
    if Len > 0 then
    begin
      Value.Position := 0;
      Stream.CopyFrom(Value, Len);
      Value.Position := 0;
    end;
  end else begin
    Stream.Read(Len, 4);
    if Len > $FFFFFF then
      RaiseNamedError('SubStream Size Limit exceded');
    Value.Size := Len;
    if Len > 0 then
    begin
      Value.Position := 0;
      Value.CopyFrom(Stream, Len);
      Value.Position := 0;
    end;
  end;
end;

procedure TWcsDataObject.StreamStringsValue(const Stream: TStream;
  const WithObjects: Boolean; const Value: TStrings;
  const Writing: Boolean; const Limit: integer = MaxInt);
var
  i,c : integer;
  S: String;
  O: Integer;
begin
  if Writing then
  begin
    c := Value.Count;
    Stream.Write(c, 4);
    for i := 0 to c-1 do
    begin
      S := Value[i];
      StreamStrValue(Stream, S, True);
      if WithObjects then
      begin
        O := Longint(Value.Objects[i]);
        Stream.Write(O, 4);
      end;
    end;
  end else begin
    Stream.Read (c, 4);
    if c > Limit then
      RaiseNamedError('DataObject SubList Count Limit exceded');
    for i := 0 to c-1 do
    begin
      StreamStrValue(Stream, S, False);
      if WithObjects then
        Stream.Read (O, 4)
      else
        O := 0;
      Value.AddObject(S, TObject(O));
    end;
  end;
end;
}

procedure TWcsDataObject.StreamResolve;
begin
// Nothing to do by default
end;

procedure TWcsDataObject.StreamIn(const Stream: TStream);
var
  DoResetEdit: Boolean;
  ProtectedStream : Boolean;
  Streamer: TWStreamer;
begin
  Clear(True);
  DoResetEdit:= EditNoModify;
  try
    if Stream.Read(fReadStreamVersion, 4) < 4 then
      raise exception.Create('Stream too short to contain Object');
    ProtectedStream := (fReadStreamVersion and $80000000) <> 0;
    fReadStreamVersion := fReadStreamVersion and $7FFFFFFF;
    if ProtectedStream then
    begin
      Streamer:= TWStreamer.CreateTStreamInput(Stream);
      try
        Streamer.StreamObject(StreamValues);
        Streamer.StreamEnd;
      finally
        Streamer.Free;
      end;
    end else begin
      StreamValuesOLD(Stream);
    end;
    StreamResolve;
  finally
    if DoResetEdit then  ResetEdit;
  end;
end;

procedure TWcsDataObject.StreamStrValueOLD(const Stream: TStream; var Value: String);
var
  Len : integer;
begin
  Stream.Read(Len, 4);
  if Len > fStringLimit then  RaiseNamedError('String Size Limit exceded');
  SetLength(Value, Len);
  if Len > 0 then  Stream.Read(Value[1], Len);
end;

procedure TWcsDataObject.StreamSubValueListOLD(const Stream: TStream;
  const Values: TList; const Limit: integer = MaxInt);
var
  i,c : integer;
  NewObject: TWcsDataObject;
begin
  Stream.Read (c, 4);
  if c > Limit then
    RaiseNamedError('SubValueList Count Limit exceded');
  for i := 0 to c-1 do
  begin
    NewObject:= CreateSubData;
    Values.Add(NewObject);
    NewObject.StreamIn(Stream);
  end;  
end;

procedure TWcsDataObject.StreamSubObject(const Streamer: TWStreamer; const AnObject: TWcsDataObject);
begin
  if Streamer.Writing then
  begin
    Streamer.StreamLongint(AnObject.fWriteStreamVersion);
    Streamer.StreamObject(AnObject.StreamValues);
  end else begin
    AnObject.EditNoModify;
    try
      Streamer.StreamLongint(AnObject.fReadStreamVersion);
      Streamer.StreamObject(AnObject.StreamValues);
      AnObject.StreamResolve;
    finally
      AnObject.ResetEdit;
    end;
  end;
end;

procedure TWcsDataObject.StreamSubValues(const Streamer: TWStreamer; var AnObject: TObject);
var
  NewObject: TWcsDataObject;
begin
  if Streamer.Writing then
  begin
    NewObject := TWcsDataObject(AnObject);
    Streamer.StreamLongint(NewObject.fWriteStreamVersion);
    Streamer.StreamObject(NewObject.StreamValues);
  end else begin
    NewObject:= CreateSubData;
    AnObject := NewObject;
    NewObject.EditNoModify;
    try
      Streamer.StreamLongint(NewObject.fReadStreamVersion);
      Streamer.StreamObject(NewObject.StreamValues);
      NewObject.StreamResolve;
    finally
      NewObject.ResetEdit;
    end;
  end;
end;

procedure TWcsDataObject.StreamOut(const Stream: TStream);
var
  ProtectedStream : Longint;
  Streamer: TWStreamer;
begin
  ProtectedStream := fWriteStreamVersion or $80000000;
  Stream.Write(ProtectedStream, 4);
  Streamer:= TWStreamer.CreateTStreamOutput(Stream);
  try
    Streamer.StreamObject(StreamValues);
    Streamer.StreamEnd;
  finally
    Streamer.Free;
  end;
end;

procedure TWcsDataObject.Backup;
begin
  fOldValues.Clear;
  StreamOut(fOldValues);
end;

procedure TWcsDataObject.Restore;
begin
  if Modified then
  begin
    fOldValues.Position := 0;
    StreamIn(fOldValues);
    Modified := False;
  end;
end;

procedure TWcsDataObject.Cancel;
begin
  case fEditState of
    wes_EditModify :
      if fModified then Restore;
    wes_EditNoCancel :
      if fModified then RaiseNamedError('Cancel not allowed!');
  end;
  SetEditState(wes_None);
end;

procedure TWcsDataObject.Post;
begin
  case fEditState of
    wes_EditModify, wes_EditNoCancel :
      if fModified then
      begin
        Modified := False;
        fUnsaved := True;
        if Assigned(fParentData) then fParentData.ChildPosted(Self);
        if Assigned(fOnAfterPost) then fOnAfterPost(Self);
      end;
  end;
  SetEditState(wes_None);
end;

procedure TWcsDataObject.ResetEdit;
begin
  fOldValues.Clear;
  fUnsaved := False;
  fModified := False;
  fEditState := wes_None;
end;

procedure TWcsDataObject.Edit;
begin
  SetEditState(wes_EditModify);
end;

procedure TWcsDataObject.EditNoCancel;
begin
  SetEditState(wes_EditNoCancel);
end;

function TWcsDataObject.EditNoModify: Boolean;
begin
  Result := fEditState = wes_None;
  if Result then SetEditState(wes_EditNoModify);
end;

function TWcsDataObject.GetEditing: Boolean;
begin
  Result := fEditState <> wes_None;
end;


procedure TWcsDataObject.SetEditState(const Value: TWcsEditState);
begin
  if fEditState <> Value then
  begin
    if (Value <> wes_None) and fReadOnly then
      RaiseNamedError('Data object is read only');
    if (Value <> wes_None) and (fEditState <> wes_None) then
      RaiseNamedError('Data object cannot change edit mode whilst editing');
    fEditState := Value;
    if Value = wes_EditModify then Backup;
    if Assigned(fParentData) then fParentData.ChildStateChange(Self);
    if Assigned(fOnStateChange) then fOnStateChange(Self);
  end;
end;

procedure TWcsDataObject.SetUnsaved(const Value: Boolean);
begin
  fUnsaved := Value;
end;

procedure TWcsDataObject.CreateSubClasses;
begin
// For descendants
end;

procedure TWcsDataObject.FreeSubClasses;
begin
// For descendants
end;

procedure TWcsDataObject.Init;
begin
// For descendants
end;

procedure TWcsDataObject.Assign(const Source: TWcsDataObject);
var
  SaveValues: TMemoryStream;
  Values: TMemoryStream;
  EditMode: TWCSEditState;
begin
  CheckEditing;
  EditMode := fEditState;
  Values:= TMemoryStream.Create;
  if EditMode = wes_EditModify then
  begin
    SaveValues:= fOldValues;
    fOldValues:= TMemoryStream.Create;
  end;
  try
    if EditMode = wes_EditModify then
      StreamOut(SaveValues);
    Source.StreamOut(Values);
    Values.Position := 0;
    StreamIn(Values);
    Modified := True;
  finally
    fEditState := EditMode;
    Values.Free;
    if EditMode = wes_EditModify then
    begin
      fOldValues.Free;
      fOldValues := SaveValues;
    end;
  end;
end;

function TWcsDataObject.CreateSubData: TWCSDataObject;
begin
  Result := Nil;
end;

procedure TWcsDataObject.ChildPosted(const ChildData: TWCSDataObject);
begin
  Modified := True;
end;

procedure TWcsDataObject.ChildStateChange(const ChildData: TWCSDataObject);
begin
// if the child goes into edit mode, so should we, unless already there!
  if fEditState = wes_None then
    case ChildData.fEditState of
      wes_EditModify   : Edit;
      wes_EditNoCancel : EditNoCancel;
      wes_EditNoModify : EditNoModify;
    end;
end;

procedure TWcsDataObject.LinkChild(const ChildData: TWCSDataObject);
begin
  ChildData.fParentData:= Self;
end;

function TWcsDataObject.GetDataObjName: String;
begin
  Result := 'Un-named Data Object';
end;

function TWcsDataObject.GetCurrStreamVersion(const Writing: Boolean): integer;
begin
  if Writing then Result := fWriteStreamVersion else Result := fReadStreamVersion;
end;

procedure TWcsDataObject.UpdateSubStreams;
begin
  fReadStreamVersion := fWriteStreamVersion;
end;

procedure TWcsDataObject.CorrectReadStreamVersion(
  const CorrectVersion: integer);
begin
  fReadStreamVersion := CorrectVersion;
end;

end.


